from typing import Tuple

import torch

from ..containers import KeyEnforcer
from ..hooks import AlignerPlusCHook, RTNHook
from ..utils.common_functions import check_domain
from .base_adapter import BaseGCAdapter
from .utils import with_opt


class Aligner(BaseGCAdapter):
    """
    Extends [BaseGCAdapter][pytorch_adapt.adapters.base_adapter.BaseGCAdapter]
    and wraps [AlignerPlusCHook][pytorch_adapt.hooks.AlignerPlusCHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C"]```|
    |optimizers|```["G", "C"]```|
    """

    hook_cls = AlignerPlusCHook

    def init_hook(self, hook_kwargs):
        opts = with_opt(list(self.optimizers.keys()))
        self.hook = self.hook_cls(opts, **hook_kwargs)


class RTN(Aligner):
    """
    Extends [Aligner][pytorch_adapt.adapters.Aligner]
    and wraps [RTNHook][pytorch_adapt.hooks.RTNHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C", "residual_model"]```|
    |optimizers|```["G", "C", "residual_model"]```|
    |misc|```["feature_combiner"]```|
    """

    hook_cls = RTNHook

    def inference_default(self, x, domain=None) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Arguments:
            x: The input to the model
            domain: If 0, ```logits = residual_model(C(G(x)))```.
                Otherwise, ```logits = C(G(x))```.
        Returns:
            Features and logits
        """
        domain = check_domain(self, domain)
        features, logits = super().inference_default(x, domain)
        if domain == 0:
            return features, self.models["residual_model"](logits)
        return features, logits

    def get_key_enforcer(self) -> KeyEnforcer:
        ke = super().get_key_enforcer()
        ke.requirements["models"].append("residual_model")
        ke.requirements["optimizers"].append("residual_model")
        ke.requirements["misc"] = ["feature_combiner"]
        return ke



from ..containers import KeyEnforcer, MultipleContainers, Optimizers
from ..hooks import ClassifierHook, FinetunerHook
from .base_adapter import BaseGCAdapter
from .utils import default_optimizer_tuple, with_opt


class Classifier(BaseGCAdapter):
    """
    Extends [BaseGCAdapter][pytorch_adapt.adapters.base_adapter.BaseGCAdapter]
    and wraps [ClassifierHook][pytorch_adapt.hooks.ClassifierHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C"]```|
    |optimizers|```["G", "C"]```|
    """

    hook_cls = ClassifierHook

    def init_hook(self, hook_kwargs):
        opts = with_opt(list(self.optimizers.keys()))
        self.hook = self.hook_cls(opts=opts, **hook_kwargs)


class Finetuner(Classifier):
    """
    Extends [Classifier][pytorch_adapt.adapters.Classifier]
    and wraps [FinetunerHook][pytorch_adapt.hooks.FinetunerHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C"]```|
    |optimizers|```["C"]```|
    """

    hook_cls = FinetunerHook

    def get_default_containers(self) -> MultipleContainers:
        optimizers = Optimizers(default_optimizer_tuple(), keys=["C"])
        return MultipleContainers(optimizers=optimizers)

    def get_key_enforcer(self) -> KeyEnforcer:
        ke = super().get_key_enforcer()
        ke.requirements["optimizers"].remove("G")
        return ke



from ...utils import common_functions as c_f
from .. import utils as f_utils
from .ignite import Ignite


# This is for the case where
# G outputs preds
# C is Identity()
# In other words, the "features" are softmaxed logits
class IgnitePredsAsFeatures(Ignite):
    def get_collector_step(self, inference):
        def collector_step(engine, batch):
            batch = c_f.batch_to_device(batch, self.device)
            return f_utils.collector_step(
                inference, batch, f_utils.create_output_dict_preds_as_features
            )

        return collector_step



from typing import Tuple

import torch

from ..hooks import SymNetsHook
from ..utils.common_functions import check_domain
from .base_adapter import BaseGCAdapter
from .utils import with_opt


class SymNets(BaseGCAdapter):
    """
    Extends [BaseGCAdapter][pytorch_adapt.adapters.base_adapter.BaseGCAdapter]
    and wraps [SymNetsHook][pytorch_adapt.hooks.SymNetsHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C"]```|
    |optimizers|```["G", "C"]```|

    The C model must output a list of logits: ```[logits1, logits2]```.
    """

    hook_cls = SymNetsHook

    def inference_default(self, x, domain) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Arguments:
            x: The input to the model
            domain: 0 for the source domain, 1 for the target domain.
        Returns:
            Features and logits, where ```logits = C(features)[domain]```.
        """
        domain = check_domain(self, domain)
        features = self.models["G"](x)
        logits = self.models["C"](features)[domain]
        return features, logits

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(
            g_opts=with_opt(["G"]), c_opts=with_opt(["C"]), **hook_kwargs
        )



import torch


def create_output_dict(features, logits):
    return {
        "features": features,
        "logits": logits,
        "preds": torch.softmax(logits, dim=1),
    }


def create_output_dict_preds_as_features(features, logits):
    if not torch.allclose(features, logits):
        raise ValueError(f"features and logits should be equal")
    return {
        "features": features,
        "preds": features,
    }


def extract_data(batch):
    data = {"_".join(k.split("_")[1:]): v for k, v in batch.items()}
    if len(data) != len(batch):
        raise KeyError(
            f"Batch should have only one domain, but it has keys: {batch.keys()}"
        )
    return data


def collector_step(inference, batch, output_dict_fn):
    data = extract_data(batch)
    with torch.no_grad():
        features, logits = inference(data["imgs"], domain=data["domain"])
    output = output_dict_fn(features, logits)
    labels_key = "labels"
    if labels_key in data:
        output[labels_key] = data[labels_key]
    return output


def filter_datasets(datasets, validator):
    return {
        k: v for k, v in datasets.items() if k in ["train"] + validator.required_data
    }



from typing import Tuple

import torch

from ..hooks import MCDHook
from .base_adapter import BaseGCAdapter
from .utils import with_opt


class MCD(BaseGCAdapter):
    """
    Extends [BaseGCAdapter][pytorch_adapt.adapters.base_adapter.BaseGCAdapter]
    and wraps [MCDHook][pytorch_adapt.hooks.MCDHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C"]```|
    |optimizers|```["G", "C"]```|

    The C model must output a list of logits, where each list element
    corresponds with a separate classifier. Usually the number of
    classifiers is 2, so C should output ```[logits1, logits2]```.
    """

    hook_cls = MCDHook

    def inference_default(self, x, domain=None) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Returns:
            Features and logits, where ```logits = sum(C(features))```.
        """
        features = self.models["G"](x)
        logits_list = self.models["C"](features)
        logits = sum(logits_list)
        return features, logits

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(
            g_opts=with_opt(["G"]), c_opts=with_opt(["C"]), **hook_kwargs
        )



import copy
from typing import Tuple

import torch

from ..containers import KeyEnforcer, MultipleContainers, Optimizers
from ..hooks import ADDAHook
from ..utils.common_functions import check_domain
from .base_adapter import BaseAdapter
from .utils import default_optimizer_tuple, with_opt


class ADDA(BaseAdapter):
    """
    Extends [BaseAdapter][pytorch_adapt.adapters.BaseAdapter]
    and wraps [ADDAHook][pytorch_adapt.hooks.ADDAHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C", "D"]```|
    |optimizers|```["D", "T"]```|

    The target model ("T") is created during initialization by deep-copying the G model.
    """

    hook_cls = ADDAHook

    def inference_default(self, x, domain) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Arguments:
            x: The input to the model
            domain: If 0, then ```features = G(x)```
                Otherwise ```features = T(x)```.
        Returns:
            Features and logits
        """
        domain = check_domain(self, domain)
        fe = "G" if domain == 0 else "T"
        features = self.models[fe](x)
        logits = self.models["C"](features)
        return features, logits

    def get_default_containers(self) -> MultipleContainers:
        """
        Returns:
            The default set of containers. This
            will create an Adam optimizer with lr 0.0001 for
            the T and D models.
        """
        optimizers = Optimizers(default_optimizer_tuple(), keys=["T", "D"])
        return MultipleContainers(optimizers=optimizers)

    def get_key_enforcer(self) -> KeyEnforcer:
        return KeyEnforcer(
            models=["G", "C", "D", "T"],
            optimizers=["D", "T"],
        )

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(
            d_opts=with_opt(["D"]), g_opts=with_opt(["T"]), **hook_kwargs
        )

    def init_containers_and_check_keys(self):
        self.containers["models"]["T"] = copy.deepcopy(self.containers["models"]["G"])
        super().init_containers_and_check_keys()



from .classifier import Classifier
from .mnist import MNISTFeatures
from .utils import download_weights


def mnistG(pretrained=False, progress=True, **kwargs):
    model = MNISTFeatures(**kwargs)
    url = "https://cornell.box.com/shared/static/tdx0ts24e273j7mf3r2ox7a12xh4fdfy"
    return download_weights(model, url, pretrained, progress)


def mnistC(num_classes=10, in_size=1200, h=256, pretrained=False, progress=True):
    model = Classifier(num_classes=num_classes, in_size=in_size, h=h)
    url = "https://cornell.box.com/shared/static/j4zrogronmievq1csulrkai7zjm27gcq"
    return download_weights(model, url, pretrained, progress)



from collections.abc import MutableMapping

import torch


def default_optimizer_tuple():
    return (torch.optim.Adam, {"lr": 0.0001})


def with_opt(x):
    suffix = "_opt"
    if isinstance(x, str):
        return f"{x}{suffix}"
    if isinstance(x, list):
        return [with_opt(y) for y in x]
    if isinstance(x, MutableMapping):
        return {with_opt(k): v for k, v in x.items()}



from torch.hub import load_state_dict_from_url


def download_weights(model, url, pretrained, progress):
    if pretrained:
        state_dict = load_state_dict_from_url(url, progress=progress)
        model.load_state_dict(state_dict)
    return model
